(* Mainframe *)
(* The EDOS project                                                 *)
(* Copyright (C)2000-2006 Berke Durak                               *)
(* Released under the GNU Lesser General Public License version 2.1 *)

open Unix;;
open Log;;
open Html;;
open Xml;;
open Http;;
open Debug;;
open Db;;
open Body_properties;;
open Command;;
module C = Configuration;;

exception Error of string;;

let bundles = [
  Cmd_login.bundle;
  Cmd_query.bundle
];;

(*** processor *)
class processor ~(config : Configuration.config) ~db () =
  let _ = Globals.init ~config ~db () in
  let global = !Globals.global in
  let sessions = global#sessions in
  let garbage_interval = Configuration.get_float ~default:60.0 config "users.garbage_interval" in
  (*** builtin_style_sheet *)
  let builtin_style_sheet = "\
    body { font-family: helvetica, sans-serif; }\n\
    div.query { font-family: courier, sans-serif; \n\
                background-color: #ffff00;\n\
                color: #000000; }\n\
    div.query span.highlight { background-color: #ff0000; }\n\
    "
  in
  (* ***)
  (*** build_head *)
  let build_head title =
    { title = title;
      author = "Anla HTTP Daemon";
      charset = UTF8;
      style_sheet = Some("stylesheet") }
  in
  (* ***)
  (*** bundles *)
  let bundles =
    List.map
      begin fun bundle ->
        bundle db config
      end
      bundles
  in
  (* ***)
  (*** command_table *)
  let command_table =
    List.fold_left
      begin fun res bundle ->
        bundle#table @ res
      end
      []
      bundles
  in
  (* ***)
  object(self)
    val mutable last_garbage = 0.0
    method name = "simple"
    method init = ()
    method shutdown =
      Log.info "Shutting down HAL 3000 bundle processors.";
      List.iter (fun bundle -> bundle#shutdown) bundles;
      global#shutdown;
      Log.info "Shutting down HAL 3000"
    (*** initializer *)
    initializer
      Log.info "Starting HAL 3000"
    (* ***)
    (*** style_sheet *)
    method private style_sheet =
      try
        Http.File(Configuration.get_string config "interface.stylesheet")
      with
      | Configuration.Key_not_found _ -> Http.String(builtin_style_sheet)
    (* ***)
    (*** decorate *)
    method private decorate form cmd_name title body =
      (*** navlines *)
      let navlines ?(hr=false) cls =
        let f x y =
          if cmd_name = x then
            Span("highlight", Anchor(x, y))
          else
            Span("normal", Anchor(x, y))
        in
        let sinfo = self#get_session form in
        Div("navlines "^cls,
            [ if hr then HR else Nop;
              if sinfo = None then
                f Link.login (T "Login")
              else
                f (Link.slink ?sinfo Link.logout) (T "Logout");
              if sinfo <> None then f (Link.slink ?sinfo Link.query)   (T "Query")   else Nop;
              if sinfo <> None then f (Link.slink ?sinfo Link.console) (T "Console") else Nop;
              if sinfo <> None then f (Link.slink ?sinfo Link.who)     (T "Who")     else Nop;
              f (Link.slink ?sinfo Link.credits) (T "Credits");
              Span("normal", Anchor("http://www.edos-project.org/", T "The EDOS project"));
              Span("normal", Anchor("http://ara.edos-project.org/", T "Ara"));
              Span("normal", Anchor("http://brion.inria.fr/anla/",  T "Anla"));
              Span("normal", Anchor("http://gallium.inria.fr/~durak/dql.html", T "DQL documentation"));
            ])
      in
      (* ***)
      Seq[
          Div("head",
            [
              Div("text", [ navlines "upper";
                            Div("title", [Span("left",H(1,T title))]) ]);
              Div("logo", [ With(["id","edos-logo"], Anchor("http://www.edos-project.org/", Nop)) ])
            ]);
          body;
          navlines ~hr:true "lower"]
    (* ***)
    (*** do_html_command *)
    method private do_html_command form cmd_name responder =
      let generate ?(properties=[]) (title, body) =
        let body = self#decorate form cmd_name title body in
        let doc = { head = build_head title;
                    body = body, properties }
        in
        Http.Okay(Http.Text_Html, Http.Html doc)
      in
      try
        generate
          begin
            match responder with
            | Anonymous resp -> resp form
            | Logged_in resp -> self#with_session form resp
          end
      with
      | Body_properties(title, body, properties) -> generate ~properties (title, body)
    (* ***)
    (*** do_raw_command *)
    method private do_raw_command form cmd_name responder =
      match responder with
      | Anonymous resp -> resp form
      | Logged_in resp ->
          try
            match self#get_session form with
            | None -> raise (Error "Not logged in")
            | Some(user, s_info) ->
                Users.touch ~sessions ~user ();
                resp ~session:(user, s_info) form
          with
          | Users.No_session -> raise (Error "Session expired or invalid session ID")
    (* ***)
    (*** get_session *)
    method private get_session form =
      try
        let sid = Form.get form "sid" in
        let (user, s_info) = Users.get_session_info ~sessions ~sid () in
        Some(user, s_info)
      with
      | Not_found|Users.No_session -> None
    (* ***)
    (*** with_session *)
    method private with_session form f =
      try
        match self#get_session form with
        | None ->
            "Error",
            P(T "You must be logged in.")
        | Some(user, s_info) ->
            Users.touch ~sessions ~user ();
            f ~session:(user, s_info) form
      with
      | Users.No_session -> "Error", P(T "Your session has expired or you have an invalid session ID.")
    (* ***)
    (*** private *)
    method private do_garbage =
      let t = Unix.gettimeofday () in
      if t -. last_garbage > garbage_interval then
        begin
          Log.info "Garbage collecting users.";
          last_garbage <- t;
          Users.garbage_collect ~sessions ()
        end
    (* ***)
    (*** process_question *)
    method process question =
      self#do_garbage;
      let rec command ?(remaining=5) form x =
        if remaining = 0 then
          raise (Error "Too many redirects")
        else
          try
            match try Some(List.assoc x command_table) with Not_found -> None
            with
            | Some(Html responder) -> self#do_html_command form x responder
            | Some(Raw responder)  -> self#do_raw_command form x responder
            | None -> Http.Error(Http.Document_not_found, x)
          with
          | Command.Redirect(form, x) ->
              Log.info (sf "Redirect to %s %s" x (Form.encode_form form));
              command ~remaining:(remaining - 1) form x
      in
      match question.Http.request with
      | Http.Get x ->
        begin
          let allowed_files =
            Configuration.get_list
              ~default:(Configuration.List []) Configuration.to_string config
              "interface.allowed_files" 
          in
          match question.Http.path with
          | [] -> command question.form "index"
          (*| ["ls"] ->
              begin
                let path = Form.get question.Http.form "path" in
                Log.info (sf "Asking for path %S" path);
                try
                  let d = Sys.readdir path in
                  let x = 
                    N("response",[],[
                        N("directory", [],
                          (N("entry", ["path", (Filename.dirname path)], [])) ::
                          (List.map
                            (fun x -> N("entry", ["path", (Filename.concat path x)], [])) (Array.to_list d)))])
                  in
                  Http.Okay(Http.Text_Xml, Xml x)
                with
                | x -> Http.Error(Http.Internal_server_error, sf "Exception %s" (Printexc.to_string x))
              end*)
          | ["weather"] ->
              let x =
                 N("response",[],[
                   N("comment",[], [A (S "How ya doin' ?")]);
                   N("random", [], [A (I (Random.int 1_000_000))])])
              in
              Http.Okay(Http.Text_Xml, Xml x)
          | ["stylesheet"] -> Http.Okay(Http.Text_Css, self#style_sheet)
          | ["css";x] ->
              if List.mem x allowed_files then
                Http.Okay(Http.Text_Css, File x)
              else
                Http.Error(Http.Forbidden, String.concat "/" question.Http.path)
          | ["html";x] ->
              if List.mem x allowed_files then
                Http.Okay(Http.Text_Html, File x)
              else
                Http.Error(Http.Forbidden, String.concat "/" question.Http.path)
          | ["javascript";x] ->
              if List.mem x allowed_files then
                Http.Okay(Http.Script_Javascript, File x)
              else
                Http.Error(Http.Forbidden, String.concat "/" question.Http.path)
          | ["image";("jpeg"|"png" as fmt);x] ->
              if List.mem x allowed_files then
                Http.Okay(Http.Image(match fmt with "jpeg"->Http.JPEG|"png"->Http.PNG|_->assert false), Http.File x)
              else
                Http.Error(Http.Forbidden, String.concat "/" question.Http.path)
          | [x] -> command question.form x
          | _ -> Http.Error(Http.Document_not_found, x)
        end
      | _ -> Http.Error(Http.Not_implemented, "Method not implemented")
    (* ***)
  end
;;
(* ***)
(*** main *)
let _ =
  Httpd.main ~processor:
    begin fun config ->
      let get_date k =
        try
          Some(Util.iso_8601_of_string (Configuration.get_string config k))
        with
        | Configuration.Key_not_found _ -> None
      in
      let g f x = try Some(f x) with Configuration.Key_not_found _ -> None in
      let p = {
        Db.par_cache_database = Configuration.get_bool config "db.cache";
        Db.par_reload         = Configuration.get_bool config "db.reload";
        Db.par_marshal_file   = Configuration.get_string config "db.marshal_file";
        Db.par_start_date     = get_date "db.start_date";
        Db.par_end_date       = get_date "db.end_date";
        Db.par_base           = Configuration.get_string config "db.base";
        Db.par_architecture   = Configuration.get_string config "db.architecture";
        Db.par_progress       = Progress.dummy }
      in
      (*** Adjust GC parameters *)
      let gcc = Gc.get () in
      gcc.Gc.space_overhead <- Configuration.get_int config ~default:gcc.Gc.space_overhead "control.gc.space_overhead";
      gcc.Gc.max_overhead   <- Configuration.get_int config ~default:gcc.Gc.space_overhead "control.gc.max_overhead";
      Log.info (sf "Adjusting GC: space_overhead=%d, max_overhead=%d" gcc.Gc.space_overhead gcc.Gc.max_overhead);
      Gc.set gcc;
      (* ***)
      Log.info "Loading database...";
      Db.with_database p
        begin fun db ->
          Log.info "Database loaded, starting";
          new processor ~config ~db ()
        end
    end
  ()
;;
(* ***)
