(* Users *)
(* The EDOS project                                                 *)
(* Copyright (C)2000-2006 Berke Durak                               *)
(* Released under the GNU Lesser General Public License version 2.1 *)

open Io;;
module G = Generic_io;;

let sf = Printf.sprintf;;

type session_id = string

type user = {
  u_login : string;
  u_name : string;
  u_salt : int;
  u_md5 : string;
  u_privilege_level : int;
  mutable u_last_login : float;
  mutable u_session : session_id option;
}

class type ['session_info] session_manager = 
  object
    method open_session : user -> session_id -> 'session_info
    method close_session : 'session_info -> unit
  end
;;

type 'session_info sessions = {
  s_users : (string, user) Hashtbl.t;
  s_sessions : (session_id, string * 'session_info) Hashtbl.t;
  mutable s_number : int;
  s_manager : 'session_info session_manager;
  s_mutex : Mutex.t;
  s_timeout : float
}

exception Already_exists;;
exception Unknown_user;;
exception Bad_password;;
exception No_session;;

let _ = Random.self_init ();;

let io_user =
  io_record
  [
    "login",
    (fun x i -> { x with u_login = read io_string i }),
    (fun x -> write io_string x.u_login);
    "name",
    (fun x i -> { x with u_name = read io_string i }),
    (fun x -> write io_string x.u_name);
    "salt",
    (fun x i -> { x with u_salt = read io_int i }),
    (fun x -> write io_int x.u_salt);
    "md5",
    (fun x i -> { x with u_md5 = read io_string i }),
    (fun x -> write io_string x.u_md5);
    "privilege_level",
    (fun x i -> { x with u_privilege_level = read io_int i }),
    (fun x -> write io_int x.u_privilege_level) ]
  { u_login = "";
    u_name = "";
    u_salt = 0;
    u_md5 = "";
    u_privilege_level = 0;
    u_last_login = 0.0;
    u_session = None }
;;

(*** load *)
let load ~manager ~fn ?(timeout=7200.0) () =
  let h =
    try
      G.load_from_file (io_hashtbl io_string io_user) fn;
    with
    | x ->
        Log.error (sf "Error loading users file: %s." (Printexc.to_string x));
        Hashtbl.create 256
  in
  { s_users = h;
    s_number = 0;
    s_manager = manager;
    s_sessions = Hashtbl.create 16;
    s_mutex = Mutex.create ();
    s_timeout = timeout }
;;
(* ***)
(*** digest *)
let digest ~salt ~login ~password =
  Hex.of_string
    (Digest.string (Printf.sprintf "%d#%S#%S#%d" salt login password salt))
;;
(* ***)
let x = Util.wrap;;
(*** under_lock *)
let under_lock : 'a sessions -> (unit -> 'b) -> 'b =
  fun sessions f ->
  Mutex.lock sessions.s_mutex;
  Util.wrap
    sessions.s_mutex
    begin fun m -> Mutex.unlock m end
    begin fun _ -> f () end
;;
(* ***)
(*** save *)
let save ~fn ~sessions () =
  under_lock sessions
    begin fun () ->
      G.save_to_file (io_hashtbl io_string io_user) sessions.s_users fn
    end
;;
(* ***)
(*** who *)
let who ~sessions () =
  under_lock sessions
    begin fun () ->
      let res = ref [] in
      Hashtbl.iter
        begin fun _ u ->
          res := (u.u_login, u.u_name, u.u_session <> None) :: !res
        end
        sessions.s_users;
      !res
    end
(* ***)
(*** register *)
let register ~sessions ~login ~name ~password ~privilege_level () =
  under_lock sessions
    begin fun () ->
      if Hashtbl.mem sessions.s_users login then raise Already_exists;
      let salt = Random.bits () in
      let md5 = digest ~salt ~login ~password in
      Hashtbl.add
        sessions.s_users
        login
        { u_login = login;
          u_name = name;
          u_salt = salt;
          u_md5 = md5;
          u_privilege_level = privilege_level;
          u_last_login = 0.0;
          u_session = None }
    end
;;
(* ***)
(*** check_login_availability *)
let check_login_availability ~sessions ~login () =
  under_lock sessions
    begin fun () ->
      if Hashtbl.mem sessions.s_users login then false
      else true
    end
;;
(* ***)
(*** logout_no_lock *)
let logout_no_lock ~sessions ~login () =
  try
    let u = Hashtbl.find sessions.s_users login in
    match u.u_session with
    | None -> ()
    | Some s_id ->
        let (_, s_info) = Hashtbl.find sessions.s_sessions s_id in
        sessions.s_manager#close_session s_info;
        u.u_session <- None;
        Hashtbl.remove sessions.s_sessions s_id;
        sessions.s_number <- sessions.s_number - 1
  with
  | Not_found -> raise Unknown_user
;;
(*** logout *)
let logout ~sessions ~login () =
  under_lock sessions
    begin fun () ->
      logout_no_lock ~sessions ~login ()
    end
;;
(* ***)
(* ***)
(*** find_new_session_id *)
let find_new_session_id ~sessions =
  let rec loop i =
    if i > sessions.s_number * 100 then failwith "Can't find valid session ID";
    let sid = Random_source.get_safe_string 16 in
    if Hashtbl.mem sessions.s_sessions sid then
      loop (i + 1)
    else
      Hex.of_string sid
  in
  loop 0
;;
(* ***)
(*** touch *)
let touch ~sessions ~user () =
  user.u_last_login <- Unix.gettimeofday ()
;;
(* ***)
(*** garbage_collect *)
let garbage_collect ~sessions () =
  under_lock sessions
    begin fun () ->
      let t = Unix.gettimeofday () in
      Hashtbl.iter
        begin fun login u ->
          if u.u_session <> None then
            begin
              let d = t -. u.u_last_login in
              if d > sessions.s_timeout then
                begin
                  Log.info (sf "User %S last activity since %f seconds more than %f, timed out" login d sessions.s_timeout);
                  logout_no_lock ~sessions ~login ()
                end
              else
                Log.debug (sf "User %S last activity since %f seconds less than %f, OK" login d sessions.s_timeout);
            end
        end
        sessions.s_users
    end
;;
(* ***)
(*** login *)
let login ~sessions ~login ~password () =
  match
    under_lock sessions
      begin fun () ->
        try
          Some(Hashtbl.find sessions.s_users login)
        with
        | Not_found -> None
      end
  with
  | None -> raise Unknown_user
  | Some u ->
    (* Validate user name and password *)
    let d = digest ~salt:u.u_salt ~login ~password in
    if d = u.u_md5 then
      begin
        (* User validated. *)
        if u.u_session <> None then logout ~sessions ~login ();
        let sid = find_new_session_id ~sessions in
        let s_info = sessions.s_manager#open_session u sid in
        u.u_session <- Some sid;
        u.u_last_login <- Unix.gettimeofday ();
        under_lock sessions
          begin fun () ->
            Hashtbl.add sessions.s_sessions sid (u.u_login, s_info);
            sessions.s_number <- 1 + sessions.s_number;
          end;
        sid
      end
    else
      raise Bad_password
;;
(* ***)
(*** get_session_info *)
let get_session_info ~sessions ~sid () =
  under_lock sessions
  begin fun () ->
    try
      let (login, s_info) = Hashtbl.find sessions.s_sessions sid in
      let u = Hashtbl.find sessions.s_users login in
      (u, s_info)
    with
    | Not_found -> raise No_session
  end
;;
(* ***)
