(* Waterway *)
(* Copyright 2005-2007 Berke DURAK, INRIA Rocquencourt and The EDOS Project. *)
(* Released under the GNU LGPL version 2. *)

open Rapids
module R = Rapids
module DBI = Dosebase.In
;;

type waterway = File of string | Directory of string | Dose of string
type specification = (liquid * waterway) list
;;

(*** specification_of_string *)
let specification_of_string u =
  let sl = Util.split_at ',' u in
  List.map
    begin fun v ->
      let (lq,fn) = Util.split_once_at ((=) ':') v in
      match lq with
      | "deb"     -> (Debian, File fn)
      | "dose"    -> (Debian, Dose fn)
      | "rpm"     -> (RPM, File fn)
      | "doserpm" -> (RPM, Dose fn)
      | _         -> invalid_arg (Printf.sprintf "Bad waterway specification %S" lq)
    end
    sl
;;
(* ***)
(*** mode_of_specification *)
let mode_of_specification sl =
  let mode = ref None in
  List.iter
    begin fun (lq, _) ->
      match (lq, !mode) with
      | (Debian, (None|Some Debian)) -> mode := Some Debian
      | (RPM,    (None|Some RPM)) -> mode := Some RPM
      | (_,_) -> invalid_arg "RPM and Debian do not mix"
    end
    sl;
  !mode
;;
(* ***)
(*** load_from_dosebase *)
let load_from_dosebase db (progress : Progress.indicator) base =
  let dbi = DBI.open_in base in
  let lifetimes = ref [] in
  progress#start;
  let count = ref 0 in
  DBI.iter
    ~package:
      begin fun ~unit_name ~version ~architecture package ->
        incr count;
        progress#display (!count,0);
        let nk = Debparser.napkin_of_fields (DBI.get_field_first dbi package) in
        ignore (add_package db nk)
      end
    ~lifetime:
      begin fun ~archive ~unit_name ~version ~architecture lifetime ->
        incr count;
        progress#display (!count,0);
        lifetimes := (unit_name, version, architecture, archive, lifetime) :: !lifetimes;
      end
    dbi;
  let unit_index = get_unit_index db in
  let architecture_index = get_architecture_index db in
  let version_index = get_version_index db in
	let release_index = get_release_index db in
  let package_index = get_package_index db in
  List.iter
    begin fun (unit_name, version, architecture, archive_name, lifetime) ->
      let unit_id          = Unit_index.search unit_index unit_name in
      let architecture_id  = Architecture_index.search architecture_index architecture in
			let v, r             = Rapids.split_version version in
      let version_id       = Version_index.get_id (Version_index.search version_index v) in
      let release_id       = Release_index.get_id (Release_index.search release_index r) in
      let package_id       = Package_index.search1 package_index (unit_id, version_id, release_id, architecture_id) in
      let archive_id       = add_archive db archive_name in
      let archive          = get_archive db archive_id in
      add_packages_to_archive
        db
        archive
        (Package_set.singleton package_id)
        lifetime
    end
    !lifetimes;
  DBI.close_in dbi
;;
(* ***)
(*** load_from_debian *)
let load_from_debian db (progress : Progress.indicator) fn =
  let count = ref 0 in
  progress#set_label (Printf.sprintf "Loading %S" fn);
  let ic = open_in fn in
  let ps = Debparser.from_channel
    ~tick:
        begin fun () ->
          incr count;
          progress#display (!count,0);
        end
    ~warn:
        begin fun w ->
          progress#warn (Printf.sprintf "Warning: %s" w)
        end
    ic
  in
  Stream.iter
    begin fun nk ->
      let _ = add_package db nk in
      ()
    end
    ps;
  close_in ic
;;
(* ***)
(*** adjust_comparator *)
let adjust_comparator db sl =
  match mode_of_specification sl with
  | None -> ()
  | Some Debian -> 
		begin
			match get_liquid db with
				None -> (set_liquid db Debian;
				Version_order.set_comparator Debver.compare;
				Release_order.set_comparator (fun c1 c2 ->
					match (c1, c2) with
					| (None, None) -> 0
					| (None, _) -> 1
					| (_, None) -> -1
					| (Some x1, Some y1) -> Debver.compare x1 y1
				))
			| Some Debian -> ()
			| Some RPM -> invalid_arg "Debian and RPM do not match"
		end
  | Some RPM -> 
		begin
			match get_liquid db with
				None -> (set_liquid db RPM;
				Version_order.set_comparator Rpmver.compare;
				Release_order.set_comparator (fun c1 c2 ->
					match (c1, c2) with
					| (None, None) -> 0
					| (None, _) -> 1
					| (_, None) -> -1
					| (Some x1, Some y1) -> Rpmver.compare x1 y1
				))
			| Some RPM -> ()
			| Some Debian -> invalid_arg "Debian and RPM do not match"
		end
;;
(*** merge *)
let merge db sl =
  adjust_comparator db sl; 
  List.iter
    begin function
      | (_,      Dose base) -> load_from_dosebase db Progress.dummy base
      | (Debian, File fn)   -> load_from_debian db Progress.dummy fn
      | (_,      _)         -> invalid_arg "Unsupported waterway"
    end
    sl
;;
(* ***)
(*** perform *)
let perform sl handle =
  let db = create_database () in
  let progress = new Progress.indicator ~decimation:1000 ~label:"Loading" ~channel:stdout () in
  progress#start;
  List.iter
    begin function
      | (_,      Dose base) -> load_from_dosebase db progress base
      | (Debian, File fn)   -> load_from_debian db progress fn
      | (_,      _)         -> invalid_arg "Unsupported waterway"
    end
    sl;
  progress#finish;
  handle db
;;
(* ***)
(*** with_database *)
(* let with_database sl handle =
  adjust_comparator sl;
  perform sl handle
;;  *)
(* ***)
